from harbor.test import BaseHarborTestClass
from harbor.tasks.configsmanagement import EnableConfigTask
from harbor.tasks.configsmanagement import DisableConfigTask
from harbor.tasks.configsmanagement import ListConfigsTask


class ConfigsManagementTest(BaseHarborTestClass):

    def test_functional_configs_can_be_listed_enabled_and_disabled(self):
        """Test listing, enabling and disabling
        """

        self.assertIn('conf/infrastructure.service-discovery.yml  Yes', self.execute_task(ListConfigsTask(), args={}))

        with self.subTest('Disable service discovery'):
            disable_task = DisableConfigTask()
            disable_task.git_commands_suffix = '|| true'
            disable_task.git_mv_command = 'mv'

            self.execute_task(disable_task, args={'--name': 'conf/infrastructure.service-discovery.yml'})
            self.assertRegex(self.execute_task(ListConfigsTask(), args={}),
                             'conf/infrastructure.service-discovery.yml.*No')

        with self.subTest('Enable service discovery'):
            enable_task = EnableConfigTask()
            enable_task.git_commands_suffix = '|| true'
            enable_task.git_mv_command = 'mv'

            self.execute_task(enable_task, args={'--name': 'conf/infrastructure.service-discovery.yml'})
            self.assertRegex(self.execute_task(ListConfigsTask(), args={}),
                             'conf/infrastructure.service-discovery.yml.*Yes')



from argparse import ArgumentParser
from subprocess import CalledProcessError
from rkd.contract import ExecutionContext
from .base import BaseProfileSupportingTask
from .base import UpdateStrategy


class StartTask(BaseProfileSupportingTask):
    """Create and start containers
    """
    
    def configure_argparse(self, parser: ArgumentParser):
        super().configure_argparse(parser)
        parser.add_argument('--strategy', '-s',
                            help='Enforce an update strategy (optional)',
                            default='auto',
                            type=UpdateStrategy, choices=list(UpdateStrategy))
        parser.add_argument('--remove-previous-images', action='store_true',
                            help='Remove previous images if service had changed docker image')

    def get_name(self) -> str:
        return ':start'

    def run(self, context: ExecutionContext) -> bool:
        services = self.get_matching_services(context)
        strategy = context.get_arg('--strategy')
        result = True

        with self.hooks_executed(context, 'start'):
            for service in services:
                self.io().h2('Starting "%s" (%i instances)...' % (service.get_name(), service.get_desired_replicas_count()))

                try:
                    self.rkd(
                        [
                            '--no-ui',
                            ':harbor:service:up',
                            service.get_name(),
                            '--remove-previous-images' if context.get_arg('--remove-previous-images') else '',
                            ('--strategy=%s' % strategy) if strategy else ''
                        ],
                        capture=not self.io().is_log_level_at_least('info')
                    )

                    self.io().success_msg('Service "%s" was started' % service.get_name())

                except CalledProcessError as e:
                    self.io().err(str(e))
                    self.io().error_msg('Cannot start service "%s"' % service.get_name())
                    result = False

                self.io().print_opt_line()

        return result


class StopTask(BaseProfileSupportingTask):
    """Stop running containers (preserving the order - the gateway should be turned off first)
    """

    def get_name(self) -> str:
        return ':stop'

    def run(self, ctx: ExecutionContext) -> bool:
        services = self.get_matching_services(ctx)

        for service in services:
            self.io().info('Stopping "%s"' % service.get_name())
            self.containers(ctx).stop(service.get_name())

        return True


class RestartTask(BaseProfileSupportingTask):
    """Restart running containers
    """

    def get_name(self) -> str:
        return ':restart'

    def run(self, ctx: ExecutionContext) -> bool:
        services = self.get_matching_services(ctx)

        for service in services:
            self.io().info('Restarting "%s"' % service.get_name())
            self.containers(ctx).restart(service.get_name())

        return True


class StopAndRemoveTask(BaseProfileSupportingTask):
    """Forcibly stop running containers and remove (keeps volumes)

    Use --with-image to delete image of each service
    """

    def get_name(self) -> str:
        return ':remove'

    def run(self, ctx: ExecutionContext) -> bool:
        services = self.get_matching_services(ctx)

        for service in services:
            self.io().info('Restarting "%s"' % service.get_name())
            self.rkd([
                ':harbor:service:rm',
                service.get_name(),
                '--with-image' if ctx.get_arg('--with-image') else ''
            ])

        return True


class PullTask(BaseProfileSupportingTask):
    """Pull images specified in containers definitions
    """

    def get_name(self) -> str:
        return ':pull'

    def run(self, ctx: ExecutionContext) -> bool:
        self.containers(ctx).pull(self.get_matching_service_names(ctx))

        return True


class UpgradeTask(BaseProfileSupportingTask):
    """Begin an upgrade procedure

    Behavior:
     1. Pull all images
     2. Pull all git repositories
     3. Upgrade services one-by-one
     4. Restart gateway
     5. Call SSL to refresh
    """

    def get_name(self) -> str:
        return ':upgrade'

    def configure_argparse(self, parser: ArgumentParser):
        super(UpgradeTask, self).configure_argparse(parser)
        parser.add_argument('--strategy', '-s',
                            help='Enforce an update strategy (optional)',
                            default='auto',
                            type=UpdateStrategy, choices=list(UpdateStrategy))
        parser.add_argument('--remove-previous-images', action='store_true',
                            help='Remove previous images if service had changed docker image')

    def run(self, context: ExecutionContext) -> bool:
        profile = context.get_arg('--profile')
        strategy = context.get_arg('--strategy')
        success = True

        with self.hooks_executed(context, 'upgrade'):
            self.rkd([
                '--no-ui',
                ':harbor:pull', '--profile=%s' % profile,

                ':harbor:start', '--profile=%s' % profile, '--strategy=%s' % strategy,
                '--remove-previous-images' if context.get_arg('--remove-previous-images') else '',

                ':harbor:gateway:reload'
            ])

        return success



from harbor.test import BaseHarborTestClass
from harbor.tasks.service import ExecTask
from harbor.service import ServiceDeclaration


class ServiceStopTaskTest(BaseHarborTestClass):
    def test_exec_can_pass_simple_commandline(self):
        """Run command inside container, do not test output as it is passed-through directly to the console"""
        service = ServiceDeclaration('alpine_3', {})

        drv = self._get_prepared_compose_driver()
        drv.up(service, capture=True)

        # assert it will not raise an error
        self.execute_task(ExecTask(), args={
            'name': 'alpine_3',
            '--extra-args': '',
            '--command': 'exit 0',
            '--shell': '/bin/sh',
            '--instance-num': None,

            # in tests its reversed
            '--no-tty': False,
            '--no-interactive': True
        })



from typing import Dict
from argparse import ArgumentParser
from tabulate import tabulate
from rkd.contract import ExecutionContext
from .base import BaseProfileSupportingTask
from ..service import ServiceDeclaration


class ListDefinedServices(BaseProfileSupportingTask):
    """Lists all defined containers in YAML files (can be limited by --profile selector)"""

    def get_group_name(self) -> str:
        return ':harbor:service'

    def get_name(self) -> str:
        return ':list'

    def configure_argparse(self, parser: ArgumentParser):
        super().configure_argparse(parser)
        parser.add_argument('--group-by', '-g', help='Group list by: url, none. Default: none', default='none')

    def run(self, ctx: ExecutionContext) -> bool:
        group_by = ctx.get_arg('--group-by')
        services = self.get_matching_services(ctx)

        # table
        table_headers = ['Priority', 'Name', 'Declared version', 'Replicas', 'URL', 'Ports',
                         'Maintenance mode', 'Update strategy', 'Watchtower']
        table_body = []

        running = self.containers(ctx).get_created_containers(only_running=True)

        for service in services:
            domains = service.get_domains()

            # GROUP-BY: list per-domain
            if group_by == 'url':
                for domain in domains:
                    self._append_to_table(table_body, service, running, domain)

                # list also services that are not under a domain (internal services or exposed via ports)
                if not domains:
                    self._append_to_table(table_body, service, running, domain='-')

            # GROUP-BY: none
            if group_by == 'none':
                self._append_to_table(table_body, service, running, domain=self._format_domains(domains))

        self.io().outln(tabulate(table_body, headers=table_headers))

        return True

    @staticmethod
    def _format_domains(domains: list):
        return "\n".join(domains)

    @staticmethod
    def _append_to_table(table_body: list, service: ServiceDeclaration,
                         running_services: Dict[str, Dict[int, bool]], domain: str):

        is_running = service.get_name() in running_services
        replicas_running = ' [%i/%i]' % (
            len(running_services[service.get_name()]) if is_running else 0,
            service.get_desired_replicas_count()
        )

        table_body.append([
            service.get_priority_number(),
            service.get_name(),
            service.get_declared_version(),
            replicas_running,
            domain,
            "\n".join(service.get_ports()),
            bool2str(service.is_using_maintenance_mode(), y='Available', n='N/A'),
            service.get_update_strategy(),
            bool2str(service.is_using_watchtower())
        ])


def bool2str(val: bool, y: str = 'Active', n: str = 'Not active'):
    return y if val else n



from subprocess import CalledProcessError
from rkd.contract import ExecutionContext
from rkd.syntax import TaskDeclaration
from harbor.test import BaseHarborTestClass
from harbor.tasks.running import UpgradeTask
from harbor.tasks.running import StopAndRemoveTask
from harbor.tasks.running import StopTask
from harbor.tasks.running import StartTask
from harbor.tasks.running import RestartTask
from harbor.exception import ProfileNotFoundException


class TestRunning(BaseHarborTestClass):
    """Set of functional tests for :start, :stop, :restart, :upgrade"""

    def test_upgrade_calls_tasks_in_order(self):
        """Assert that :upgrade calls other Harbor tasks in fixed order"""

        task = UpgradeTask()
        recorded_calls = []
        task.rkd = lambda *args, **kwags: recorded_calls.append(args)

        self.execute_task(task, args={
            '--profile': 'profile1',
            '--strategy': 'recreate',
            '--remove-previous-images': False
        })

        called_tasks_in_order = list(filter(
            lambda element: element.startswith(':'),
            recorded_calls[0][0]
        ))

        self.assertEqual([':harbor:pull', ':harbor:start', ':harbor:gateway:reload'], called_tasks_in_order)

    def test_upgrade_calls_tasks_with_proper_arguments(self):
        """Check if arguments such as --profile and --strategy are passed"""

        task = UpgradeTask()
        recorded_calls = []
        task.rkd = lambda *args, **kwags: recorded_calls.append(args)

        self.execute_task(task, args={
            '--profile': 'profile1',
            '--strategy': 'recreate',
            '--remove-previous-images': False
        })

        args = ' '.join(recorded_calls[0][0])

        self.assertIn(':harbor:pull --profile=profile1', args)
        self.assertIn(':harbor:start --profile=profile1 --strategy=recreate', args)

    def test_stop_and_remove_task_reports_invalid_profile(self):
        task = StopAndRemoveTask()
        recorded_calls = []
        task.rkd = lambda *args, **kwags: recorded_calls.append(args)

        self.assertRaises(ProfileNotFoundException, lambda: self.execute_task(task, args={
            '--profile': 'non-existing',
            '--strategy': 'recreate',
        }))

    def test_stop_and_remove_task_iterates_through_matched_profiles_and_calls_service_removal(self):
        task = StopAndRemoveTask()
        recorded_calls = []
        task.rkd = lambda *args, **kwags: recorded_calls.append(args)

        self.execute_task(task, args={
            '--profile': 'profile1',
            '--with-image': False
        })

        args = list(map(lambda call: ' '.join(call[0]).strip(), recorded_calls))
        self.assertEqual([':harbor:service:rm gateway', ':harbor:service:rm website'], args)

    def test_restart_calls_driver_restart_method_on_matched_services(self):
        """Test calls restart on expected services in expected order"""

        task = RestartTask()
        restarted_services = []

        ctx = ExecutionContext(
            TaskDeclaration(task),
            args={},
            env={}
        )
        task.containers(ctx).restart = lambda service_name, args = '': restarted_services.append(service_name)

        self.execute_task(task, args={
            '--profile': 'profile1',
            '--with-image': False
        })

        self.assertEqual(['gateway', 'website'], restarted_services)

    def test_stop_task_executes_stop_task_multiple_times(self):
        """Test that StopTask will call Driver.stop() multiple times"""

        task = StopTask()
        recorded_calls = []

        ctx = ExecutionContext(
            TaskDeclaration(task),
            args={},
            env={}
        )
        task.containers(ctx).stop = lambda service_name, args='', capture = False: recorded_calls.append(service_name)

        self.execute_task(task, args={
            '--profile': ''
        })

        self.assertIn('gateway', recorded_calls)
        self.assertIn('gateway_letsencrypt', recorded_calls)
        self.assertIn('gateway_proxy_gen', recorded_calls)

    def test_start_task_executes_tasks_startup_in_order(self):
        """Basic test to check if profile is considered, and if startup order is preserved"""

        task = StartTask()
        recorded_calls = []
        task.rkd = lambda *args, **kwags: recorded_calls.append(args)

        self.execute_task(task, args={
            '--profile': 'profile1',
            '--strategy': 'rolling',
            '--remove-previous-images': False
        })

        args = list(map(lambda call: ' '.join(call[0]).strip(), recorded_calls))
        commandline = ' '.join(args).replace('  ', ' ').strip()

        self.assertEqual(
            '--no-ui :harbor:service:up gateway --strategy=rolling --no-ui :harbor:service:up website --strategy=rolling',
            commandline
        )

    def test_start_task_on_single_failure_continues_but_returns_false_at_the_end(self):
        """Checks that even if one service was not started correctly, it will not break up whole deployment
        The next services should start normally, but the result at the end should be a 'failure'"""

        task = StartTask()
        recorded_calls = []

        def rkd_mock(*args, **kwargs):
            if len(recorded_calls) == 0:
                recorded_calls.append(args)
                raise CalledProcessError(1, 'bash')

            recorded_calls.append(args)

        task.rkd = rkd_mock

        out = self.execute_task(task, args={
            '--profile': 'profile1',
            '--strategy': 'rolling',
            '--remove-previous-images': False
        })

        self.assertIn('Cannot start service "gateway"', out)
        self.assertIn('Service "website" was started', out)
        self.assertIn('TASK_EXIT_RESULT=False', out)



from harbor.test import BaseHarborTestClass
from harbor.tasks.maintenance import MaintenanceOnTask
from harbor.tasks.maintenance import MaintenanceOffTask


class TestMaintenanceModeFeature(BaseHarborTestClass):
    """Functional tests for the maintenance feature

    Includes coverage of:
        - Tasks
        - NGINX template
    """

    def test_functional_global_maintenance_will_be_on_then_off(self):
        """Simply test global maintenance mode switch for services that have label
         `org.riotkit.useMaintenanceMode: true`"""

        drv = self._get_prepared_compose_driver()
        self.prepare_service_discovery(drv)
        self.prepare_example_service('website_with_maintenance', uses_service_discovery=True)

        # 1) maintenance on
        with self.subTest('Maintenance mode on'):
            self.execute_task(MaintenanceOnTask(), args={'--global': True, '--domain': '', '--service': ''})
            content = self.fetch_page_content('nginx-with-maintenance-mode.local')

            self.assertIn('503: Page maintenance', content)
            self.assertIn('HTTP/1.1 503', content)

        # 2) maintenance mode off
        with self.subTest('Maintenance mode off'):
            self.execute_task(MaintenanceOffTask(), args={'--global': True, '--domain': '', '--service': ''})
            content = self.fetch_page_content('nginx-with-maintenance-mode.local')

            self.assertNotIn('503: Page maintenance', content)
            self.assertIn('HTTP/1.1 200', content)

    def test_functional_global_maintenance_mode_not_affects_service_that_has_maintenance_off(self):
        """Verify that service that does not have `org.riotkit.useMaintenanceMode` label should not display maintenance
        page, even if global maintenance is on"""

        drv = self._get_prepared_compose_driver()
        self.prepare_service_discovery(drv)
        self.prepare_example_service('website', uses_service_discovery=True)

        # 1) maintenance on, then check - should not be under maintenance as service has maintenance off in definition
        self.execute_task(MaintenanceOnTask(), args={'--global': True, '--domain': '', '--service': ''})

        self.assertIn('HTTP/1.1 200', self.fetch_page_content('nginx.local'))

    def test_functional_maintenance_mode_for_service(self):
        """Verify that if we want to put a SERVICE in maintenance mode, then all DOMAINS of that SERVICE will be
        in maintenance"""

        drv = self._get_prepared_compose_driver()
        self.prepare_service_discovery(drv)
        self.prepare_example_service('website_with_multiple_domains', uses_service_discovery=True)

        # 1) maintenance on
        with self.subTest('Turning maintenance on'):
            self.execute_task(MaintenanceOnTask(), args={
                '--global': False, '--domain': '', '--service': 'website_with_multiple_domains'
            })

            self.assertIn('HTTP/1.1 503', self.fetch_page_content('web1.local'))
            self.assertIn('HTTP/1.1 503', self.fetch_page_content('web2.local'))
            self.assertIn('HTTP/1.1 503', self.fetch_page_content('web3.local'))

        # 2) maintenance off
        with self.subTest('Turning off the maintenance'):
            self.execute_task(MaintenanceOffTask(), args={
                '--global': False, '--domain': '', '--service': 'website_with_multiple_domains'
            })

            self.assertIn('HTTP/1.1 200', self.fetch_page_content('web1.local'))
            self.assertIn('HTTP/1.1 200', self.fetch_page_content('web2.local'))
            self.assertIn('HTTP/1.1 200', self.fetch_page_content('web3.local'))

    def test_functional_maintenance_mode_for_single_domain_under_service_that_has_multiple_domains(self):
        """Test that we can set a maintenance for domain web1.local, while web2.local stays normal
        WHERE both domains are under SAME service"""

        drv = self._get_prepared_compose_driver()
        self.prepare_service_discovery(drv)
        self.prepare_example_service('website_with_multiple_domains', uses_service_discovery=True)

        # 1) maintenance on
        self.execute_task(MaintenanceOnTask(), args={
            '--global': False, '--domain': 'web1.local', '--service': ''
        })

        self.assertIn('HTTP/1.1 503', self.fetch_page_content('web1.local'))
        self.assertIn('HTTP/1.1 200', self.fetch_page_content('web2.local'))

    def test_validation_of_domain_parameter(self):
        """Try to set a maintenance mode for domain not defined in YAML"""

        drv = self._get_prepared_compose_driver()
        self.prepare_service_discovery(drv)
        self.prepare_example_service('website_with_multiple_domains', uses_service_discovery=True)

        out = self.execute_task(MaintenanceOnTask(), args={
            '--global': False, '--domain': 'invalid_domain.local', '--service': ''
        })

        self.assertIn('Domain is not valid', out)

    def test_validation_of_service_parameter(self):
        """Try to set a maintenance mode for service that is not defined in YAML"""

        drv = self._get_prepared_compose_driver()
        self.prepare_service_discovery(drv)
        self.prepare_example_service('website_with_multiple_domains', uses_service_discovery=True)

        out = self.execute_task(MaintenanceOnTask(), args={
            '--global': False, '--domain': '', '--service': 'non_existing_service'
        })

        self.assertIn('Service "non_existing_service" was not defined', out)



import os
import pkg_resources
from rkd.contract import ExecutionContext
from rkd.standardlib import CreateStructureTask

HARBOR_PATH = os.path.dirname(os.path.realpath(__file__)) + '/..'


class CreateHarborStructureTask(CreateStructureTask):
    """Create a new Harbor project structure"""

    def get_group_name(self) -> str:
        return ':harbor:create'

    def get_name(self) -> str:
        return ':project'

    def on_startup(self, ctx: ExecutionContext) -> None:
        self.io().info_msg('Creating Harbor structure...')

    def on_files_copy(self, ctx: ExecutionContext) -> None:
        """Copies files, if project was not created yet"""

        self.sh('cp -prfT %s/project ./' % HARBOR_PATH)

    def on_requirements_txt_write(self, ctx: ExecutionContext) -> None:
        """Apply Ansible to requirements.txt"""

        self.rkd([':file:line-in-file',
                  'requirements.txt',
                  '--regexp="ansible(.*)"',
                  '--insert="ansible>=2.8"'
                  ])

        self.rkd([':file:line-in-file',
                  'requirements.txt',
                  '--regexp="rkd-harbor(.*)"',
                  '--insert="rkd-harbor%s"' % self.get_harbor_version_matcher()
                  ])

    @staticmethod
    def get_harbor_version_matcher() -> str:
        harbor_version = pkg_resources.get_distribution("rkd-harbor").version

        return '==' + harbor_version


    def print_success_msg(self, ctx: ExecutionContext) -> None:
        super().print_success_msg(ctx)
        self.io().print_line()
        self.io().success_msg("Harbor successfully installed on bootstrapped RKD project, enjoy - " +
                              "RiotKit tech collective.")




"""Services declarations operations - docker-compose YAML represented as model
"""

import os
from typing import List
from typing import Optional
from typing import Dict
from traceback import format_exc
from rkd.inputoutput import IO
from .expressions import safe_eval
from .exception import ProfileNotFoundException
from .exception import ServiceNotFoundInYaml
from .exception import ServiceNotFoundInYamlLookedByCriteria

DEFAULT_SELECTOR = 'service is not None'  # passes all containers
BOOLEANS = ['true', 'TRUE', 'True', True]


class ServiceDeclaration(object):
    """Model of a parsed service declaration from YAML"""

    _name: str
    _definition: dict

    def __init__(self, name: str, definition: dict):
        self._name = name
        self._definition = definition

    def get_name(self) -> str:
        return self._name

    def get_definition(self) -> dict:
        return self._definition

    def get_domains(self) -> list:
        try:
            return str(self.get_definition()['environment']['VIRTUAL_HOST']).split(',')
        except KeyError:
            return []

    def is_using_watchtower(self) -> bool:
        try:
            return self.get_definition()['labels']['com.centurylinklabs.watchtower.enable'] in BOOLEANS
        except KeyError:
            return False

    def is_using_maintenance_mode(self) -> bool:
        try:
            return self.get_definition()['labels']['org.riotkit.useMaintenanceMode'] in BOOLEANS
        except KeyError:
            return False

    def get_ports(self) -> list:
        try:
            return self.get_definition()['ports']
        except KeyError:
            return []

    def get_desired_replicas_count(self) -> int:
        try:
            return int(self.get_definition()['labels']['org.riotkit.replicas'])
        except KeyError:
            return 1

    def get_update_strategy(self, default: str = 'compose') -> str:
        try:
            return str(self.get_definition()['labels']['org.riotkit.updateStrategy'])
        except KeyError:
            return default

    def get_priority_number(self):
        try:
            return int(self.get_definition()['labels']['org.riotkit.priority'])
        except KeyError:
            return 1000

    def get_image(self):
        try:
            return str(self.get_definition()['image'])
        except KeyError:
            return '_docker_build_local:latest'

    def get_declared_version(self):
        try:
            return str(self.get_definition()['image'].split(':')[1])
        except KeyError:
            return 'latest (build)'
        except IndexError:
            return 'latest'

    def has_domain(self, domain: str):
        return domain in self.get_domains()


class ServiceSelector(object):
    """Acts as a service filter. Simple reduce() implementation"""

    _selector: str
    _io: IO

    def __init__(self, selector: str, io: IO):
        self._selector = selector
        self._io = io

    def is_service_matching(self, definition: dict, name: str) -> bool:
        """Asks the profile filter - is service of a given definition and name matching?"""

        try:
            return safe_eval(self._selector, {'service': definition, 'name': name})
        except Exception:
            self._io.errln(format_exc())
            self._io.error_msg('Exception raised, while attempting to evaluate --profile selector')
            return False

    def find_matching_services(self, services: dict) -> List[ServiceDeclaration]:
        """Find names of matching services by current Service Selector"""
        matched = []

        for name, definition in services.items():
            if self.is_service_matching(definition, name):
                matched.append(ServiceDeclaration(name, definition))

        matched.sort(key=lambda declaration: declaration.get_priority_number())

        return matched


class ProfileLoader(object):
    """Parses profiles from ./apps/profiles
    """

    _io: IO
    _apps_path: str

    def __init__(self, io: IO, apps_path: str):
        self._io = io
        self._apps_path = apps_path

    def load_profile(self, name: str) -> ServiceSelector:
        if name == '' or name is None:
            return ServiceSelector(DEFAULT_SELECTOR, self._io)

        profile_path = self._apps_path + '/profile/%s.profile.py' % name

        if not os.path.isfile(profile_path):
            raise ProfileNotFoundException(profile_path)

        return self.load_profile_from_path(profile_path)

    def load_profile_from_path(self, path: str) -> ServiceSelector:
        with open(path, 'r') as f:
            content = f.read()

            return ServiceSelector(content, self._io)


class ServiceLocator(object):
    """Declarations repository"""

    _services: Dict[str, ServiceDeclaration]

    def __init__(self, services: dict):
        self._services = {}

        for name, yaml_dict in services.items():
            self._services[name] = ServiceDeclaration(name, yaml_dict)

    def get_by_name(self, name: str) -> Optional[ServiceDeclaration]:
        try:
            return self._services[name]
        except KeyError:
            raise ServiceNotFoundInYaml(name)

    def find_by_domain(self, domain: str) -> Optional[ServiceDeclaration]:
        for service in self._services.values():
            if service.has_domain(domain):
                return service

        raise ServiceNotFoundInYamlLookedByCriteria('has domain "%s"' % domain)



from argparse import ArgumentParser
from .base import HarborBaseTask
from rkd.contract import ExecutionContext


class DumpComposeArguments(HarborBaseTask):
    """Shows arguments that Harbor passes to docker-compose all the time"""

    def get_group_name(self) -> str:
        return ':harbor:diagnostic'

    def get_name(self) -> str:
        return ':dump-compose-args'

    def run(self, context: ExecutionContext) -> bool:
        try:
            self.io().outln(self.containers(context).get_compose_args())

        except AttributeError as e:
            self.io().error_msg('Cannot retrieve compose arguments, possibly not using docker-compose. ' +
                                'Details: %s' % str(e))
            return False

        return True


class ListContainersTask(HarborBaseTask):
    """List all containers
    """

    def get_group_name(self) -> str:
        return ':harbor:diagnostic:compose'

    def get_name(self) -> str:
        return ':ps'

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--quiet', '-q', help='Only display IDs', action='store_true')
        parser.add_argument('--all', '-a', help='Show all containers, including stopped', action='store_true')

    def run(self, ctx: ExecutionContext) -> bool:
        params = []

        if ctx.get_arg('--quiet'):
            params.append('--quiet')

        if ctx.get_arg('--all'):
            params.append('--all')

        self.containers(ctx).ps(params)
        return True


class DumpComposeConfigTask(HarborBaseTask):
    """Dump all compose YAMLs as one YAML"""

    def get_group_name(self) -> str:
        return ':harbor:diagnostic:compose'

    def get_name(self) -> str:
        return ':config'

    def run(self, context: ExecutionContext) -> bool:
        self.containers(context).compose(['config'])

        return True



from harbor.test import BaseHarborTestClass
from harbor.tasks.service import ServiceStopTask
from harbor.service import ServiceDeclaration


class ServiceStopTaskTest(BaseHarborTestClass):
    def test_functional_service_is_stopped(self):
        """Start service, then try to bring down, leave the images"""
        service = ServiceDeclaration('alpine_3', {})

        drv = self._get_prepared_compose_driver()
        drv.up(service, capture=True)

        self.execute_task(ServiceStopTask(), args={
            'name': 'alpine_3',
            '--extra-args': ''
        })

        self.assertContainerIsNotRunning('alpine_3', drv)



from harbor.test import BaseHarborTestClass
from harbor.tasks.service import WaitForServiceTask
from harbor.service import ServiceDeclaration
from harbor.exception import ServiceNotFoundInYaml


class WaitForServiceTaskTest(BaseHarborTestClass):
    def test_functional_waits_for_service_without_healthcheck(self):
        """Expect that the task will wait for container to be up, but will not check its health check if healthcheck
        was not defined"""

        drv = self._get_prepared_compose_driver()
        drv.up(ServiceDeclaration('alpine_3', {}), capture=True)
        drv.stop('alpine_3', capture=True)

        out = self.execute_task(WaitForServiceTask(), args={
            'name': 'alpine_3',
            '--timeout': 1,
            '--instance': 1
        })

        self.assertIn('Instance has no healthcheck defined!', out)

    def test_functional_timeout_is_raised(self):
        """Expects that the timeout will be raised, as it is very low in test"""

        drv = self._get_prepared_compose_driver()
        drv.up(ServiceDeclaration('alpine_3_health_check', {}), capture=True)

        out = self.execute_task(WaitForServiceTask(), args={
            'name': 'alpine_3_health_check',
            '--timeout': 0,
            '--instance': 1
        })

        self.assertIn('Timeout of 0s reached.', out)

    def test_functional_container_is_healthy_after_few_seconds_using_health_check_invoking_while_container_is_starting(self):
        """Expects that the container will become health within few seconds,
        even if docker daemon is showing 'starting'"""

        drv = self._get_prepared_compose_driver()
        drv.rm(ServiceDeclaration('alpine_3_health_check', {}), capture=True)
        drv.up(ServiceDeclaration('alpine_3_health_check', {}), capture=True)

        out = self.execute_task(WaitForServiceTask(), args={
            'name': 'alpine_3_health_check',
            '--timeout': 10,
            '--instance': 1
        })

        self.assertIn('Service healthy after', out)

    def test_will_not_wait_for_invalid_service(self):
        """Expects that the container will become health within few seconds,
        even if docker daemon is showing 'starting'"""

        self.assertRaises(ServiceNotFoundInYaml, lambda: self.execute_task(WaitForServiceTask(), args={
                                'name': 'not_existing_service',
                                '--timeout': 10,
                                '--instance': 1
                            })
                          )





from harbor.test import BaseHarborTestClass
from harbor.tasks.service import ServiceUpTask
from harbor.service import ServiceDeclaration


class ServiceUpTaskTest(BaseHarborTestClass):
    def test_functional_service_is_getting_recreated(self):
        """Test that service is recreated"""

        drv = self._get_prepared_compose_driver()
        drv.up(ServiceDeclaration('website', {}), capture=True)
        drv.stop('website', capture=True)

        for i in range(1, 2):
            out = self.execute_task(ServiceUpTask(), args={
                'name': 'website',
                '--strategy': 'recreate',
                '--remove-previous-images': False,
                '--extra-args': ''
            })

            self.assertIn('Recreating env_simple_website_1', out)

    def test_functional_service_is_rolling_updated(self):
        """Test that service rolling update is performed"""

        drv = self._get_prepared_compose_driver()
        drv.rm(ServiceDeclaration('website', {}), capture=True)
        drv.up(ServiceDeclaration('website', {}), capture=True)

        # start gateway (it needs to be later checked if it is stopped an then started)
        drv.up(ServiceDeclaration('gateway_proxy_gen', {}), capture=True)

        out = self.execute_task(ServiceUpTask(), args={
            'name': 'website',
            '--strategy': 'rolling',
            '--remove-previous-images': False,
            '--extra-args': ''
        })

        self.assertIn('Stopping env_simple_gateway_proxy_gen_1', out,
                      msg='Expected the rolling strategy will stop proxy_gen')
        self.assertIn('Starting env_simple_gateway_proxy_gen_1', out,
                      msg='Expected the rolling strategy will start proxy_gen')

        running_containers = self.get_containers_state(drv)

        self.assertIn('env_simple_website_2', running_containers,
                      msg='Expected that a new instance would be in place of old one')
        self.assertNotIn('env_simple_website_1', running_containers,
                         msg='Expected that the old container will be turned off')

    def test_functional_compose_method_is_executed(self):
        """Test that the service is not force-recreated by 'compose' method"""

        drv = self._get_prepared_compose_driver()
        drv.rm(ServiceDeclaration('website', {}), capture=True)
        drv.up(ServiceDeclaration('website', {}), capture=True)

        out = self.execute_task(ServiceUpTask(), args={
            'name': 'website',
            '--strategy': 'compose',
            '--remove-previous-images': False,
            '--extra-args': '',
            '--dont-recreate': False
        })

        self.assertIn('env_simple_website_1 is up-to-date', out)

    def test_strategy_is_automatically_selected_from_service_definition(self):
        """Strategy if not defined in YAML will be fallback to default: 'compose'"""

        drv = self._get_prepared_compose_driver()
        drv.rm(ServiceDeclaration('website', {}), capture=True)

        out = self.execute_task(ServiceUpTask(), args={
            'name': 'website',
            '--strategy': 'auto',
            '--remove-previous-images': False,
            '--extra-args': '',
            '--dont-recreate': False
        })

        self.assertIn('Performing "compose"', out)

    def test_invalid_strategy_raises_error(self):
        """Tests that enum is validated"""

        drv = self._get_prepared_compose_driver()
        drv.rm(ServiceDeclaration('website', {}), capture=True)

        out = self.execute_task(ServiceUpTask(), args={
            'name': 'website',
            '--strategy': 'invalid-strategy-name',
            '--remove-previous-images': False,
            '--extra-args': '',
            '--dont-recreate': False
        })

        self.assertIn('Invalid strategy selected: invalid-strategy-name', out)



from harbor.test import BaseHarborTestClass
from harbor.service import ServiceSelector
from rkd.inputoutput import BufferedSystemIO
from rkd.inputoutput import IO


class ServiceSelectorTest(BaseHarborTestClass):
    def _provide_test_data(self) -> dict:
        return {
            'web_phillyabc': {
                'image': 'nginx:1.19',
                'labels': {
                    'org.riotkit.type': 'abc',
                    'org.riotkit.country': 'USA',
                    'org.riotkit.priority': 150
                }
            },
            'web_abc_international': {
                'image': 'nginx:1.19',
                'labels': {
                    'org.riotkit.type': 'abc',
                    'org.riotkit.priority': 100
                }
            },
            'web_iwa_ait': {
                'image': 'nginx:1.19',
                'labels': {
                    'org.riotkit.type': 'workers-union',
                    'org.riotkit.priority': 200
                }
            }
        }

    def test_find_matching_services_by_labels(self):
        """Test that services can be found by labels"""

        io = IO()
        selector = ServiceSelector(
            '"org.riotkit.type" in service["labels"] and service["labels"]["org.riotkit.type"] == "abc"', io
        )

        names = list(map(lambda service: service.get_name(),
                         selector.find_matching_services(self._provide_test_data())))

        self.assertEqual(['web_abc_international', 'web_phillyabc'], names)

    def test_find_matching_services_by_names(self):
        io = IO()
        selector = ServiceSelector('name.startswith("web")', io)

        names = list(map(lambda service: service.get_name(),
                         selector.find_matching_services(self._provide_test_data())))

        self.assertEqual(['web_abc_international', 'web_phillyabc', 'web_iwa_ait'], names)

    def test_find_matching_skips_services_on_syntax_error_but_raises_error_in_log(self):
        io = BufferedSystemIO()
        selector = ServiceSelector('service["labels"]["org.riotkit.country"] != ""', io)

        names = list(map(lambda service: service.get_name(),
                         selector.find_matching_services(self._provide_test_data())))

        self.assertEqual(['web_phillyabc'], names)
        self.assertIn("KeyError: 'org.riotkit.country'", io.get_value())
        self.assertIn("Exception raised, while attempting to evaluate --profile selector", io.get_value())

    def test_finding_matching_services_considers_priority_ascending(self):
        io = IO()
        selector = ServiceSelector('True', io)

        names = list(map(lambda service: service.get_name(),
                         selector.find_matching_services(self._provide_test_data())))

        with self.subTest('With test data from data provider'):
            self.assertEqual(['web_abc_international', 'web_phillyabc', 'web_iwa_ait'], names)

        with self.subTest('With modified priority orders'):
            test_data = self._provide_test_data()
            test_data['web_iwa_ait']['labels']['org.riotkit.priority'] = 50
            test_data['web_phillyabc']['labels']['org.riotkit.priority'] = -3

            names = list(map(lambda service: service.get_name(),
                             selector.find_matching_services(test_data)))

            self.assertEqual(['web_phillyabc', 'web_iwa_ait', 'web_abc_international'], names)

    def test_finding_matching_services_accepts_1000_as_default_priority(self):
        io = IO()
        selector = ServiceSelector('True', io)

        test_data = self._provide_test_data()
        test_data['web_iwa_ait']['labels']['org.riotkit.priority'] = 900
        del test_data['web_phillyabc']['labels']['org.riotkit.priority']
        test_data['web_abc_international']['labels']['org.riotkit.priority'] = 1200

        names = list(map(lambda service: service.get_name(), selector.find_matching_services(test_data)))

        self.assertEqual(['web_iwa_ait', 'web_phillyabc', 'web_abc_international'], names)



import subprocess
from harbor.test import BaseHarborTestClass
from harbor.tasks.deployment import EnvEncryptTask


class EnvEncryptTaskTest(BaseHarborTestClass):
    def test_functional_env_is_encrypted_then_decrypted(self):
        """Verify that the structure is copied, and the files are rendered using JINJA2"""

        # 1) Encryption
        with self.subTest('Encrypted file ".env-prod" is produced'):
            self.execute_task(EnvEncryptTask(),
                              args={
                                  '--ask-vault-pass': False,
                                  '--vault-passwords': 'union',
                                  '--decrypt': False
                              },
                              env={})

            with open(self.get_test_env_subdirectory('') + '/.env-prod', 'rb') as f:
                self.assertIn('$ANSIBLE_VAULT;', f.read().decode('utf-8'))

        # before next subtest remove the original .env file contents,
        # so we can check if it will be recreated from .env-prod
        subprocess.check_call('echo "" > .env', shell=True)

        # 2) Decryption
        with self.subTest('File ".env" is decrypted back from ".env-prod"'):
            self.execute_task(EnvEncryptTask(),
                              args={
                                  '--ask-vault-pass': False,
                                  '--vault-passwords': 'union',
                                  '--decrypt': True
                              },
                              env={})

            with open(self.get_test_env_subdirectory('') + '/.env', 'rb') as f:
                content = f.read().decode('utf-8')

                self.assertNotIn('$ANSIBLE_VAULT;', content)
                self.assertIn('COMPOSE_PROJECT_NAME=env_simple', content)




import os
import yaml
import pkg_resources
from contextlib import contextmanager
from subprocess import check_output
from argparse import ArgumentParser
from abc import ABC
from abc import abstractmethod
from typing import Dict
from typing import List
from enum import Enum
from rkd.contract import ExecutionContext
from ..service import ProfileLoader
from ..service import ServiceDeclaration
from ..service import ServiceLocator
from ..driver import ComposeDriver
from ..cached_loader import CachedLoader
from ..interface import HarborTaskInterface

SCRIPT_PATH = os.path.dirname(os.path.realpath(__file__))


class UpdateStrategy(Enum):
    rolling = 'rolling'
    recreate = 'recreate'
    compose = 'compose'
    auto = 'auto'

    def __str__(self):
        return self.value


class HarborBaseTask(HarborTaskInterface):
    _compose_args: str
    is_dev_env: bool
    app_user: str
    app_group_id: int

    #
    # TaskInterface
    #
    def configure_argparse(self, parser: ArgumentParser):
        pass

    def get_group_name(self) -> str:
        return ':harbor'

    def get_declared_envs(self) -> Dict[str, str]:
        return {
            'APPS_PATH': './apps/',
            'HOOKS_PATH': './hooks.d/',
            'DATA_PATH': './data/',
            'COMPOSE_PROJECT_NAME': None,
            'DOMAIN_SUFFIX': ''
        }

    #
    # Lazy-dependencies
    #
    def services(self, ctx: ExecutionContext) -> ServiceLocator:
        return CachedLoader.cached('services', lambda: ServiceLocator(self.get_services_as_raw_dict(ctx)))

    def containers(self, ctx: ExecutionContext) -> ComposeDriver:
        return CachedLoader.cached('containers',
                                   lambda: ComposeDriver(self, ctx, self.get_project_name(ctx)))

    def profile_loader(self, ctx: ExecutionContext) -> ProfileLoader:
        """Loads profile of a ServiceSelector to filter out"""

        return ProfileLoader(self.io(), self.get_apps_path(ctx))

    #
    # Validation
    #
    @staticmethod
    def _validate_env_present() -> bool:
        return os.path.isfile('./.env')

    #
    # Configuration
    #
    @staticmethod
    def get_apps_path(ctx: ExecutionContext) -> str:
        return ctx.get_env('APPS_PATH')

    @staticmethod
    def get_hooks_dir(ctx: ExecutionContext) -> str:
        return ctx.get_env('HOOKS_PATH')

    @staticmethod
    def get_data_path(ctx: ExecutionContext) -> str:
        return ctx.get_env('DATA_PATH')

    @staticmethod
    def get_project_name(ctx: ExecutionContext) -> str:
        return ctx.get_env('COMPOSE_PROJECT_NAME')

    def get_compose_yaml_as_dict(self, ctx: ExecutionContext):
        """Return's parsed docker-compose file as one big dictionary"""

        return CachedLoader.load_compose_definition(
            lambda: yaml.load(self.containers(ctx).compose(['config'], capture=True), yaml.Loader)
        )

    def get_services_as_raw_dict(self, ctx: ExecutionContext):
        """Gets services from YAMLS"""
        parsed = self.get_compose_yaml_as_dict(ctx)

        return parsed['services'] if 'services' in parsed else {}

    @abstractmethod
    def run(self, context: ExecutionContext) -> bool:
        """Method to implement the proper logic of the task (instead of execute())
        """
        pass

    def execute(self, context: ExecutionContext) -> bool:
        """Harbor's wrapper - adds Harbor specific behavior before each task - initialization of the context
        """

        if not self._validate_env_present():
            self.io().error_msg('Missing .env file')
            return False

        self.app_user, self.app_group_id = self.detect_repository_owning_user_and_group()
        self.is_dev_env = self.detect_dev_env(context)
        project_name = context.get_env('COMPOSE_PROJECT_NAME')

        # project directory has to have owner that will manage files, that should be a deployment user
        self.io().debug('Project owner: USER=%s, GID=%i' % (self.app_user, self.app_group_id))

        if not project_name:
            self.io().error_msg('COMPOSE_PROJECT_NAME environment variable is not defined, cannot proceed')
            return False

        return self.run(context)

    @staticmethod
    def detect_repository_owning_user_and_group() -> tuple:
        """Detects user and group that should be owner of the project repository"""

        usr = os.getenv('APP_USER', '')
        grp = os.getenv('APP_GROUP_ID', '')

        if not usr and os.getenv('SUDO_USER'):
            usr = os.getenv('SUDO_USER')

        if not grp and os.getenv('SUDO_GID'):
            grp = os.getenv('SUDO_GID')

        if not usr:
            usr = os.getenv('USER')

        if not grp:
            grp = check_output(['id', '-g', usr]).decode('utf-8').strip()

        return usr, int(grp)

    @staticmethod
    def detect_dev_env(context: ExecutionContext) -> bool:
        suffix = context.get_env('DOMAIN_SUFFIX')

        return suffix.endswith('.localhost') or suffix.endswith('.xip.io')

    @contextmanager
    def hooks_executed(self, ctx: ExecutionContext, hooks_suffix: str):
        self.execute_hooks(ctx, 'pre-' + hooks_suffix)
        yield
        self.execute_hooks(ctx, 'post-' + hooks_suffix)

    def execute_hooks(self, ctx: ExecutionContext, hook_name: str):
        """Executes a hooks from directory "hooks.d", when any executed script will fail, then an exception is raised

        The error handling must be implemented in the scripts itself
        """

        hooks_dir = self.get_hooks_dir(ctx) + '/' + hook_name + '/'

        if not os.path.isdir(hooks_dir):
            self.io().debug('Hooks dir "%s" not present, skipping' % hooks_dir)
            return

        self.io().info('Executing hook scripts for action "%s"' % hook_name)

        for script in os.scandir(hooks_dir):
            script: os.DirEntry

            if script.name.split('.')[0] == '':
                continue

            self.io().info('Running hook script "%s"' % script.path)

            try:
                self.sh(script.path)
            except Exception as e:
                self.io().error_msg('Failed executing script: "%s"' % script.path)
                raise e

    def get_harbor_version(self) -> str:
        try:
            return pkg_resources.get_distribution("harbor").version
        except:
            return 'dev'

    def _silent_mkdir(self, path: str):
        try:
            os.mkdir(path)
        except FileExistsError:
            pass


class BaseProfileSupportingTask(HarborBaseTask, ABC):
    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--profile', '-p', help='Services profile', default='')

    def get_matching_services(self, ctx: ExecutionContext) -> List[ServiceDeclaration]:
        service_selector = self.profile_loader(ctx).load_profile(ctx.get_arg('--profile'))
        matched = service_selector.find_matching_services(self.get_services_as_raw_dict(ctx))

        return matched

    def get_matching_service_names(self, context: ExecutionContext) -> List[str]:
        service_names = []

        for service in self.get_matching_services(context):
            service_names.append(service.get_name())

        return service_names



from harbor.test import BaseHarborTestClass
from harbor.tasks.service import ServiceRemoveTask
from harbor.service import ServiceDeclaration


class ServiceRemoveTaskTest(BaseHarborTestClass):
    def test_functional_service_is_stopped_and_removed(self):
        """Test that service is stopped and removed, together with two instances

        1. Creates one instance + one replica
        2. Removes service
        3. Expects that two instances were removed
        """

        service = ServiceDeclaration('website', {})

        drv = self._get_prepared_compose_driver()
        drv.rm(service, capture=True)
        drv.up(service, capture=True)
        drv.scale_one_up(service)

        out = self.execute_task(ServiceRemoveTask(), args={
            'name': 'website',
            '--with-image': False,
            '--extra-args': ''
        })

        containers = self.get_containers_state(drv)

        for container_name, state in containers.items():
            if container_name.startswith('env_simple_website_'):
                self.fail('Found %s container, which should be removed by calling :harbor:service:rm')

        self.assertIn('Removing env_simple_website_1', out)
        self.assertIn('Removing env_simple_website_2', out)

    def test_functional_service_is_removed_together_with_image(self):
        service = ServiceDeclaration('alpine_3', {})

        drv = self._get_prepared_compose_driver()
        drv.up(service, capture=True)

        self.execute_task(ServiceRemoveTask(), args={
            'name': 'alpine_3',
            '--with-image': True,
            '--extra-args': ''
        })

        self.assertLocalRegistryHasNoPulledImage('alpine:3.11')



import os
from typing import List
from argparse import ArgumentParser
from abc import abstractmethod
from rkd.contract import ExecutionContext
from .base import HarborBaseTask
from ..formatting import prod_formatting
from ..exception import ServiceNotFoundInYamlLookedByCriteria
from ..exception import ServiceNotFoundInYaml


class BaseMaintenanceManagementTask(HarborBaseTask):
    def get_group_name(self) -> str:
        return ':harbor:maintenance'

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--domain', '-d', help='Domain name', default='')
        parser.add_argument('--service', '-s', help='Service name', default='')
        parser.add_argument('--global', '-g', help='Set maintenance for all domains', action='store_true')

    def run(self, context: ExecutionContext) -> bool:
        """Validate parameters and select action"""

        is_global = context.get_arg('--global')
        domain = context.get_arg('--domain')
        service = context.get_arg('--service')

        directory = self.get_data_path(context) + '/maintenance-mode'

        if not self._validate_switches(is_global, domain, service):
            self.io().error_msg('Cannot use together --global, --domain and --service switch. Pick one of them.')
            return False

        try:
            if is_global:
                return self.act([directory + '/on'], 'globally')

            elif service:
                return self.act_for_service(directory, service, context)

            elif domain:
                return self.act_for_domain(directory, domain, context)
            else:
                self.io().error_msg('Must specify --global or --domain switch')
                return False

        except PermissionError as e:
            self.io().error_msg('No write permissions. Set permissions or use sudo? %s' % str(e))
            return False

    def act_for_service(self, directory: str, service: str, ctx: ExecutionContext):
        try:
            domains = self.services(ctx).get_by_name(service).get_domains()

        except ServiceNotFoundInYaml:
            self.io().error_msg('Service "%s" was not defined' % service)
            return False

        return self.act(
            list(map(lambda domain: directory + '/%s-on' % domain, domains)),
            'for service "%s"' % service
        )

    def act_for_domain(self, directory, domain, context):
        try:
            self.services(context).find_by_domain(domain)
        except ServiceNotFoundInYamlLookedByCriteria:
            self.io().error_msg('Domain is not valid')
            return False

        return self.act([directory + '/%s-on' % domain], 'for domain "%s"' % domain)

    @abstractmethod
    def act(self, paths: List[str], subject: str) -> bool:
        pass

    def format_task_name(self, name) -> str:
        return prod_formatting(name)

    @staticmethod
    def _validate_switches(is_global, domain, service):
        switches = 0

        if is_global:
            switches += 1

        if domain:
            switches += 1

        if service:
            switches += 1

        return switches == 1


class MaintenanceOnTask(BaseMaintenanceManagementTask):
    """Turn on the maintenance mode"""

    def get_name(self) -> str:
        return ':on'

    def act(self, paths: List[str], subject: str) -> bool:
        for path in paths:
            with open(path, 'w') as f:
                f.write('We are on strike ;)')

            os.chmod(path, 0o777)

        self.io().success_msg('Maintenance mode %s is on' % subject)

        return True


class MaintenanceOffTask(BaseMaintenanceManagementTask):
    """Turn on the maintenance mode"""

    def get_name(self) -> str:
        return ':off'

    def act(self, paths: List[str], subject: str) -> bool:
        for path in paths:
            if os.path.isfile(path):
                os.unlink(path)

        self.io().success_msg('Maintenance mode %s is off' % subject)

        return True


class MaintenanceListTask:
    pass



import os
import tempfile
from harbor.test import BaseHarborTestClass
from harbor.tasks.structure import CreateHarborStructureTask


class CreateHarborStructureTaskTest(BaseHarborTestClass):

    def test_functional_installation(self):
        """Test functionally that the structure will be created properly

        NOTICE: This test has external dependencies such as PyPI
        """

        backup_dir_path = os.getcwd()

        try:
            with tempfile.TemporaryDirectory() as dir_path:
                os.chdir(dir_path)

                task = CreateHarborStructureTask()
                # do not match current dev version as it may be unreleased yet when developing
                task.get_harbor_version_matcher = lambda: ''

                self.execute_task(task, args={
                    '--commit': False,
                    '--no-venv': False
                })

                self.assertTrue(os.path.isfile(dir_path + '/.env-default'),
                                msg='Expected that dot files will be copied')

                self.assertTrue(os.path.isfile(dir_path + '/.venv/bin/activate'),
                                msg='Expected the virtualenv to be installed')

                self.assertTrue(os.path.isfile(dir_path + '/apps/README.md'),
                                msg='Expected that files in inner directories will be present')

                with open(dir_path + '/requirements.txt', 'rb') as requirements_txt:
                    content = requirements_txt.read().decode('utf-8')

                    self.assertIn('ansible>=2.8', content, msg='Expected Ansible defined in requirements.txt')
                    self.assertIn('rkd-harbor', content, msg='Expected rkd-harbor to be defined in requirements.txt')
        finally:
            os.chdir(backup_dir_path)




from harbor.test import BaseHarborTestClass
from harbor.tasks.deployment import EncryptVaultTask


class EncryptVaultTaskTest(BaseHarborTestClass):
    def test_functional_file_is_encrypted(self):
        """Test encryption"""

        with self.subTest('Encrypt the file'):
            self.execute_task(EncryptVaultTask(),
                              args={
                                  '--ask-vault-pass': False,
                                  '--vault-passwords': 'InternalWorkersAssociation',
                                  '--decrypt': False,
                                  'filename': 'README.md'
                              },
                              env={})

            with open(self.get_test_env_subdirectory('') + '/README.md', 'rb') as f:
                self.assertIn('$ANSIBLE_VAULT;', f.read().decode('utf-8'))

        with self.subTest('Decrypt the file'):
            self.execute_task(EncryptVaultTask(),
                              args={
                                  '--ask-vault-pass': False,
                                  '--vault-passwords': 'InternalWorkersAssociation',
                                  '--decrypt': True,
                                  'filename': 'README.md'
                              },
                              env={})

            with open(self.get_test_env_subdirectory('') + '/README.md', 'rb') as f:
                self.assertNotIn('$ANSIBLE_VAULT;', f.read().decode('utf-8'))





class CachedLoader(object):
    items = {}

    @classmethod
    def load_compose_definition(cls, loader: callable):
        return cls.cached('compose', loader)

    @classmethod
    def cached(cls, name: str, loader: callable):
        if name not in cls.items:
            cls.items[name] = loader()

        return cls.items[name]

    @classmethod
    def clear(cls):
        cls.items = {}



from harbor.test import BaseHarborTestClass
from harbor.tasks.service import LogsTask
from harbor.service import ServiceDeclaration


class LogsTaskTest(BaseHarborTestClass):
    def test_will_print_logs_for_container(self):
        """Run command inside container, do not test output as it is passed-through directly to the console"""
        service = ServiceDeclaration('alpine_3', {})

        drv = self._get_prepared_compose_driver()
        drv.up(service, capture=True)

        out = self.execute_task(LogsTask(), args={
            'name': 'alpine_3',
            '--instance-num': 1,
            '--follow': False,
            '--buffered': True
        })

        self.assertIn('Hello', out)
